@@ -23,6 +23,9 @@ struct ecrec {
     avuint cdir_size;
     avuint cdir_off;
     avushort comment_len;
+
+    avoff_t file_size;
+    short is_zip64;
 };
 
 #define ECREC_THIS_DISK     4
@@ -35,6 +38,46 @@ struct ecrec {
 
 #define ECREC_SIZE          22
 
+/* Z64_ECDL reads ZIP64_End_of_Central_Director_Locator */
+
+struct z64_end_of_central_dir_loc {
+    avuint ecdir_disk;
+    avuquad ecdir_off;
+    avuint total_disks;
+
+    struct ecrec *ecd;
+};
+
+#define Z64_ECDL_ECDIR_DISK   4
+#define Z64_ECDL_ECDIR_OFF    8
+#define Z64_ECDL_TOTAL_DISKS 16
+
+#define Z64_ECDL_SIZE        20
+
+#define Z64_ECD_VERSION       12
+#define Z64_ECD_NEED_VERSION  14
+#define Z64_ECD_THIS_DISK     16
+#define Z64_ECD_CDIR_DISK     20
+#define Z64_ECD_THIS_ENTRIES  24
+#define Z64_ECD_TOTAL_ENTRIES 32
+#define Z64_ECD_CDIR_SIZE     40
+#define Z64_ECD_CDIR_OFF      48
+#define Z64_ECD_SIZE          56
+
+/* Z64_ECD end of central directory record */
+
+struct z64_end_of_central_dir {
+    avushort version;
+    avushort need_version;
+    avuint   this_disk;
+    avuint   cdir_disk;
+    avuquad  this_entries;
+    avuquad  total_entries;
+    avuquad  cdir_size;
+    avuquad  cdir_off;
+};
+
+
 struct cdirentry {
     avushort version;
     avushort need_version;
@@ -42,15 +85,15 @@ struct cdirentry {
     avushort method;
     avuint mod_time;
     avuint crc;
-    avuint comp_size;
-    avuint file_size;
+    avuquad comp_size;
+    avuquad file_size;
     avushort fname_len;
     avushort extra_len;
     avushort comment_len;
     avushort start_disk;
     avushort int_attr;
     avuint attr;
-    avuint file_off;         
+    avuquad file_off;
 };
 
 #define CDIRENT_VERSION       4
@@ -77,8 +120,8 @@ struct ldirentry {
     avushort method;
     avuint mod_time;
     avuint crc;
-    avuint comp_size;
-    avuint file_size;
+    avuquad comp_size;
+    avuquad file_size;
     avushort fname_len;
     avushort extra_len;
 };
@@ -107,9 +150,14 @@ struct ldirentry {
 #define SEARCHLEN 66000
 
 #define BI(ptr, i)  ((avbyte) (ptr)[i])
+#define BI_Q(ptr, i)  ((avuquad)((avbyte) (ptr)[i]))
 #define DBYTE(ptr) (BI(ptr,0) | (BI(ptr,1)<<8))
 #define QBYTE(ptr) (BI(ptr,0) | (BI(ptr,1)<<8) | \
                    (BI(ptr,2)<<16) | (BI(ptr,3)<<24))
+#define DQBYTE(ptr) (BI_Q(ptr,0)      | (BI_Q(ptr,1)<<8) | \
+                    (BI_Q(ptr,2)<<16) | (BI_Q(ptr,3)<<24) | \
+                    (BI_Q(ptr,4)<<32) | (BI_Q(ptr,5)<<40) | \
+                    (BI_Q(ptr,6)<<48) | (BI_Q(ptr,7)<<56))
 
 struct zipnode {
     avuint crc;
@@ -184,6 +232,9 @@ static avoff_t find_ecrec(vfile *vf, long searchlen, struct ecrec *ecrec)
     ecrec->cdir_off =      QBYTE(buf+ECREC_CDIR_OFF);
     ecrec->comment_len =   DBYTE(buf+ECREC_COMMENT_LEN);
 
+    ecrec->file_size = sres;
+    ecrec->is_zip64 = 0;
+
     return bufstart;
 }
 
@@ -301,6 +352,110 @@ static void insert_zipentry(struct archive *arch, char *path,
     av_unref_obj(ent);
 }
 
+static int parse_extra_header(vfile *vf, avoff_t pos,
+                              struct cdirentry *cent,
+                              struct ldirentry *lent)
+{
+    char buf[4];
+    avoff_t end = pos + (cent ? cent->extra_len : lent->extra_len);
+
+    while (pos + 4 <= end) {
+        /* read header ID and size */
+        int res = av_pread_all(vf, buf, 4, pos);
+        avushort id;
+        int size;
+
+        if(res < 0) {
+            return res;
+        }
+
+        id = DBYTE(buf);
+        size = DBYTE(buf+2);
+
+        pos += 4;
+
+        if (id == 1 && pos + size <= end) {
+            if((cent && (avuint)(cent->file_size) == 0xffffffff) ||
+               (lent && (avuint)(lent->file_size) == 0xffffffff)) {
+                if(size >= 8) {
+                    char buf8[8];
+                    int res = av_pread_all(vf, buf8, 8, pos);
+
+                    if(res < 0) {
+                        return res;
+                    }
+
+                    if (cent) {
+                        cent->file_size = DQBYTE(buf8);
+                    } else {
+                        lent->file_size = DQBYTE(buf8);
+                    }
+
+                    size -= 8;
+                    pos += 8;
+                } else {
+                    if (cent) {
+                        cent->file_size = 0;
+                    } else {
+                        lent->file_size = 0;
+                    }
+                    pos += size;
+                }
+            }
+
+            if((cent && (avuint)(cent->comp_size) == 0xffffffff) ||
+               (lent && (avuint)(lent->comp_size) == 0xffffffff)) {
+                if(size >= 8) {
+                    char buf8[8];
+                    int res = av_pread_all(vf, buf8, 8, pos);
+
+                    if(res < 0) {
+                        return res;
+                    }
+
+                    if (cent) {
+                        cent->comp_size = DQBYTE(buf8);
+                    } else{
+                        lent->comp_size = DQBYTE(buf8);
+                    }
+
+                    size -= 8;
+                    pos += 8;
+                } else {
+                    if (cent) {
+                        cent->comp_size = 0;
+                    } else{
+                        lent->comp_size = 0;
+                    }
+                    pos += size;
+                }
+            }
+
+            if(cent && (avuint)(cent->file_off) == 0xffffffff) {
+                if(size >= 8) {
+                    char buf8[8];
+                    int res = av_pread_all(vf, buf8, 8, pos);
+
+                    if(res < 0) {
+                        return res;
+                    }
+
+                    cent->file_off = DQBYTE(buf8);
+
+                    size -= 8;
+                    pos += 8;
+                } else {
+                    pos += size;
+                }
+            }
+        } else {
+            pos += size;
+        }
+    }
+
+    return 0;
+}
+
 static avoff_t read_entry(vfile *vf, struct archive *arch, avoff_t pos,
                           struct ecrec *ecrec)
 {
@@ -342,6 +497,17 @@ static avoff_t read_entry(vfile *vf, struct archive *arch, avoff_t pos,
     }
     filename[ent.fname_len] = '\0';
 
+    if(pos + CDIRENT_SIZE + ent.fname_len + ent.extra_len +
+       ent.comment_len > ecrec->file_size) {
+        av_free(filename);
+        return -EIO;
+    }
+
+    if(parse_extra_header(vf, pos + CDIRENT_SIZE + ent.fname_len, &ent, NULL) < 0 ) {
+        av_free(filename);
+        return -EIO;
+    }
+
     insert_zipentry(arch, filename, &ent, ecrec);
     av_free(filename);
 
@@ -349,11 +515,149 @@ static avoff_t read_entry(vfile *vf, struct archive *arch, avoff_t pos,
         ent.comment_len;
 }
 
+static avoff_t find_z64_ecd(vfile *vf, struct z64_end_of_central_dir_loc *ecdl, struct z64_end_of_central_dir *z64_ecd, avoff_t pos)
+{
+    char buf[Z64_ECD_SIZE];
+    int res;
+
+    if(pos + Z64_ECD_SIZE > ecdl->ecd->file_size) {
+        return -EIO;
+    }
+
+    res = av_pread_all(vf, buf, Z64_ECD_SIZE, pos);
+    if(res < 0) {
+        return res;
+    }
+
+	if(!(buf[0] == 'P' && buf[1] == 'K' && 
+         buf[2] == 6 && buf[3] == 6)) {
+        return -EIO;
+	}
+
+    z64_ecd->version       =  DBYTE(buf+Z64_ECD_VERSION);
+    z64_ecd->need_version  =  DBYTE(buf+Z64_ECD_NEED_VERSION);
+    z64_ecd->this_disk     =  QBYTE(buf+Z64_ECD_THIS_DISK);
+    z64_ecd->cdir_disk     =  QBYTE(buf+Z64_ECD_CDIR_DISK);
+    z64_ecd->this_entries  = DQBYTE(buf+Z64_ECD_THIS_ENTRIES);
+    z64_ecd->total_entries = DQBYTE(buf+Z64_ECD_TOTAL_ENTRIES);
+    z64_ecd->cdir_size     = DQBYTE(buf+Z64_ECD_CDIR_SIZE);
+    z64_ecd->cdir_off      = DQBYTE(buf+Z64_ECD_CDIR_OFF);
+
+    return pos;
+}
+
+static avoff_t find_z64_ecdl(vfile *vf, struct ecrec *ecd, struct z64_end_of_central_dir_loc *ecdl, avoff_t pos)
+{
+    char buf[Z64_ECDL_SIZE];
+    int res;
+
+    if (pos < Z64_ECDL_SIZE) {
+        return -EIO;
+    }
+
+    pos -= Z64_ECDL_SIZE;
+
+    res = av_pread_all(vf, buf, Z64_ECDL_SIZE, pos);
+    if(res < 0) {
+        return res;
+    }
+
+	if(!(buf[0] == 'P' && buf[1] == 'K' && 
+         buf[2] == 6 && buf[3] == 7)) {
+        return -EIO;
+	}
+
+    ecdl->ecdir_disk  =    QBYTE(buf+Z64_ECDL_ECDIR_DISK);
+    ecdl->ecdir_off   =   DQBYTE(buf+Z64_ECDL_ECDIR_OFF);
+    ecdl->total_disks =    QBYTE(buf+Z64_ECDL_TOTAL_DISKS);
+
+    ecdl->ecd = ecd;
+
+    return pos;
+}
+
+static int find_and_validate_z64_ecdl(vfile *vf, struct archive *arch, struct ecrec *ecrec, avoff_t pos,
+                                      struct z64_end_of_central_dir_loc *ecdl)
+{
+    avoff_t ecdl_pos;
+
+    ecdl_pos = find_z64_ecdl(vf, ecrec, ecdl, pos);
+    if(ecdl_pos < 0) {
+        return ecdl_pos;
+    }
+
+    if(ecdl->ecdir_disk != ecdl->ecd->this_disk) {
+        return -EIO;
+    }
+
+    if (ecdl->ecdir_off > ecdl->ecd->file_size) {
+        return -EIO;
+    }
+
+    ecrec->is_zip64 = 1;
+
+    return 0;
+}
+
+static int read_zip64file(vfile *vf, struct archive *arch, struct ecrec *ecrec,
+                          struct z64_end_of_central_dir_loc *ecdl, avoff_t pos)
+{
+    struct z64_end_of_central_dir z64_ecd;
+    avoff_t extra_bytes;
+    avoff_t cdir_end;
+    avoff_t ecdir_pos;
+    avoff_t cdir_pos;
+    int nument;
+
+    ecdir_pos = ecdl->ecdir_off;
+
+    if (ecdir_pos > ecdl->ecd->file_size) {
+        return -EIO;
+    }
+
+    pos = find_z64_ecd(vf, ecdl, &z64_ecd, ecdir_pos);
+    if(pos < 0) {
+        return -EIO;
+    }
+
+    if(z64_ecd.cdir_disk != z64_ecd.this_disk) {
+        av_log(AVLOG_ERROR, "UZIP: Cannot handle multivolume archives");
+        return -EIO;
+    }
+
+    cdir_end = z64_ecd.cdir_size + z64_ecd.cdir_off;
+
+    extra_bytes = pos - cdir_end;
+    if(extra_bytes < 0) {
+        av_log(AVLOG_ERROR, "UZIP: Broken archive");
+        return -EIO;
+    }
+  
+    if(z64_ecd.cdir_off == 0 && z64_ecd.cdir_size == 0) {
+        /* Empty zipfile */
+        return 0;
+    }
+  
+    cdir_pos = z64_ecd.cdir_off + extra_bytes;
+
+    for(nument = 0; nument < z64_ecd.total_entries; nument++) {
+        if(cdir_pos >= pos) {
+            av_log(AVLOG_ERROR, "UZIP: Broken archive");
+            return -EIO;
+        }
+        cdir_pos = read_entry(vf, arch, cdir_pos, ecrec);
+        if(cdir_pos < 0) 
+            return cdir_pos;
+    }
+  
+    return 0;
+}
 
 static int read_zipfile(vfile *vf, struct archive *arch)
 {
     avoff_t ecrec_pos;
     struct ecrec ecrec;
+    struct z64_end_of_central_dir_loc ecdl;
     avoff_t extra_bytes;
     avoff_t cdir_end;
     avoff_t cdir_pos;
@@ -363,6 +667,11 @@ static int read_zipfile(vfile *vf, struct archive *arch)
     if(ecrec_pos < 0)
         return ecrec_pos;
 
+    if(find_and_validate_z64_ecdl(vf, arch, &ecrec, ecrec_pos, &ecdl) == 0) {
+        /* zip64 format */
+        return read_zip64file(vf, arch, &ecrec, &ecdl, ecrec_pos);
+    }
+
     cdir_end = ecrec.cdir_size+ecrec.cdir_off;
 
     if(ecrec.this_disk != ecrec.cdir_disk) {
@@ -388,6 +697,7 @@ static int read_zipfile(vfile *vf, struct archive *arch)
             av_log(AVLOG_ERROR, "UZIP: Broken archive");
             return -EIO;
 	}
+
 	cdir_pos = read_entry(vf, arch, cdir_pos, &ecrec);
 	if(cdir_pos < 0) 
             return cdir_pos;
@@ -476,6 +786,10 @@ static int zip_open(ventry *ve, struct archfile *fil)
     headersize = LDIRENT_SIZE + ent.fname_len + ent.extra_len;
     fil->nod->offset = offset + headersize;
 
+    if(parse_extra_header(fil->basefile, offset + LDIRENT_SIZE + ent.fname_len, NULL, &ent) < 0 ) {
+        return -EIO;
+    }
+
     if(ent.method == METHOD_DEFLATE) {
         struct zfile *zfil;
 
